﻿Imports System.Drawing.Point

Module GrahamScan

    ' Liegt Punkt C links, rechts oder auf dem Vektor AB?
    Private Function Orientation(ByVal A As Point, _
        ByVal B As Point, ByVal C As Point) As Integer

        Return (B.X - A.X) * (C.Y - A.Y) - (C.X - A.X) * (B.Y - A.Y)
    End Function

    ' Minimalen Punkt (kleinste y-Koordinate) suchen
    Private Function Less(ByVal A As Point, ByVal B As Point) As Boolean
        Return (A.Y < B.Y) OrElse ((A.Y = B.Y) And (A.X < B.X))
    End Function

    Public Function ConvexHull( _
        ByVal RawPoints As IEnumerable(Of Point)) As IEnumerable(Of Point)

        If RawPoints.Count <= 3 Then Return RawPoints

        Dim Points = New List(Of Point)(RawPoints)

        ' "Kleinsten" Punkt suchen
        Dim MinIndex = Enumerable.Range(0, Points.Count).Aggregate( _
            Function(Min, idx) If(Less(Points(idx), Points(Min)), idx, Min))
        Dim MinPoint = Points(MinIndex)

        ' Minimum an Position 0 bringen
        Call Swap(Points(0), Points(MinIndex))

        ' Restliche Punkte von links nach rechts ihrem Winkel zum minimalen 
        '  Punkt nach sortieren
        Call Points.Sort(1, Points.Count - 1, _
            New FunctionComparer(Of Point)( _
                Function(a, b) Orientation(MinPoint, a, b)))

        ' Stapel für die Elemente
        Dim Stack As New List(Of Point)

        ' Erster und zweiter Punkt bilden die Ausgangssituation
        Call Stack.Add(Points(0))
        Call Stack.Add(Points(1))

        ' Alle weiteren Punkte durchgehen 
        For i = 2 To Points.Count - 1

            ' Vorhergehende Drehungen im Uhrzeigersinn, die nicht zur Hülle 
            '  gehören, löschen
            Do Until (Stack.Count < 2) OrElse _
                (Orientation(Stack(Stack.Count - 1), _
                    Stack(Stack.Count - 2), Points(i)) > 0)

                Call Stack.RemoveAt(Stack.Count - 1)
            Loop

            ' Punkt hinzufügen
            Call Stack.Add(Points(i))
        Next

        ' Stapel ausgeben
        Return Stack
    End Function

End Module